package org.rcisoft.alibaba.faceVerify.service.impl;

import com.alibaba.fastjson.JSON;
import com.aliyun.cloudauth20190307.Client;
import com.aliyun.cloudauth20190307.models.DescribeFaceVerifyRequest;
import com.aliyun.cloudauth20190307.models.DescribeFaceVerifyResponse;
import com.aliyun.cloudauth20190307.models.InitFaceVerifyRequest;
import com.aliyun.cloudauth20190307.models.InitFaceVerifyResponse;
import com.aliyun.teaopenapi.models.Config;
import com.aliyun.teautil.models.RuntimeOptions;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import jakarta.servlet.http.HttpServletRequest;
import lombok.extern.slf4j.Slf4j;
import org.rcisoft.alibaba.faceVerify.bean.FaceVerifyAliComp;
import org.rcisoft.alibaba.faceVerify.bean.FaceVerifyRedisBean;
import org.rcisoft.alibaba.faceVerify.service.FaceVerifyService;
import org.rcisoft.business.memInfo.dao.MemInfoRepository;
import org.rcisoft.business.memInfo.entity.MemInfo;
import org.rcisoft.core.exception.CyServiceException;
import org.rcisoft.core.jwt.util.CyJwtUtil;
import org.rcisoft.core.service.CyRedisService;
import org.rcisoft.core.util.CyAESUtils;
import org.rcisoft.core.util.CyAddressUtils;
import org.rcisoft.core.util.CyUserUtil;
import org.rcisoft.core.util.TimeUtil;
import org.rcisoft.sys.rbac.user.dao.SysUserRbacRepository;
import org.rcisoft.sys.rbac.user.entity.SysUserRbac;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.context.request.RequestContextHolder;
import org.springframework.web.context.request.ServletRequestAttributes;

import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.*;

@Service
@Transactional(readOnly = true,propagation = Propagation.NOT_SUPPORTED)
@Slf4j
public class FaceVerifyServiceImpl extends ServiceImpl<MemInfoRepository, MemInfo> implements FaceVerifyService {

    @Autowired
    private FaceVerifyAliComp faceVerifyAliComp;
    @Autowired
    private MemInfoRepository memInfoRepository;
    @Autowired
    private CyJwtUtil cyJwtUtil;
    @Autowired
    private CyRedisService cyRedisServiceImpl;
    @Autowired
    private SysUserRbacRepository sysUserRbacRepository;
    /**
     * 发起认证请求
     * @param memInfo
     * @return
     */
    @Override
    public Map<String,String> initFaceVerify(HttpServletRequest httpServletRequest,MemInfo memInfo) {
//        MemInfo memInfo = memInfoRepository.getInfoByUserId(CyUserUtil.getAuthenBusinessId());
//        String idCard = CyAESUtils.decrypt(memInfo.getMemIdcard());
//        String name = CyAESUtils.decrypt(memInfo.getMemRealName());
        Object detail = cyRedisServiceImpl.get(FaceVerifyRedisBean.FACEVERIFY_USER+CyUserUtil.getAuthenBusinessId());
        if (detail!=null) {
            MemInfo info = JSON.parseObject(JSON.toJSONString(detail), MemInfo.class);
            String idCard = info.getMemIdcard();
            String name = info.getMemRealName();
            Map<String, String> map = new HashMap<>();
            InitFaceVerifyRequest request = new InitFaceVerifyRequest();
            // 场景ID+L。
            request.setSceneId(faceVerifyAliComp.getSceneId());
            // 设置商户请求的唯一标识。（请自己生成一个不超过32位数的数字及字母字符串）
            LocalDateTime now = LocalDateTime.now();
            String timestamp = now.format(DateTimeFormatter.ofPattern("yyyyMMddHHmmss"));
            String uuid = UUID.randomUUID().toString().replace("-", "");
            String outerOrderNo = timestamp + uuid.substring(0, 14);
            request.setOuterOrderNo(outerOrderNo);
            // 认证方案。
            request.setProductCode("ID_PRO");
            // 模式。LIVENESS（默认）：眨眼动作活体检测。
            request.setModel("LIVENESS");
            // 不同证件类型，取值均为IDENTITY_CARD。
            request.setCertType("IDENTITY_CARD");
            // 您的终端用户的真实姓名。
            request.setCertName(name);
            // 您的终端用户的证件号码。
            request.setCertNo(idCard);
            // MetaInfo环境参数，此参数应由前端js获取并传入。
            request.setMetaInfo(memInfo.getMetaInfo());
            // 认证结束后回跳页面的链接地址。。
            request.setReturnUrl(faceVerifyAliComp.getReturnUrl()+"?userId="+CyUserUtil.getAuthenBusinessId());
            request.setCallbackUrl(faceVerifyAliComp.getCallbackUrl());
            request.setCallbackToken(CyAESUtils.encrypt(CyUserUtil.getAuthenBusinessId()));

            InitFaceVerifyResponse response = initFaceVerifyAutoRoute(request);
            log.info("code:" + response.getBody().getCode());
            log.info("certifyId:" + response.getBody().getResultObject().getCertifyId());
            if ("200".equals(response.getBody().getCode())) {
                map.put("requestId", response.getBody().getRequestId());
                map.put("certifyId", response.getBody().getResultObject().getCertifyId());
                map.put("certifyUrl", response.getBody().getResultObject().getCertifyUrl());
            } else if ("401".equals(response.getBody().getCode())) {
                log.error("发起认证请求失败:" + response.getBody().getMessage());
                throw new CyServiceException("输入的姓名或者身份证号码有误");
            }  else if ("418".equals(response.getBody().getCode())) {
                log.error("发起认证请求失败:" + response.getBody().getMessage());
                throw new CyServiceException("当天刷脸认证次数过多，请明天再试。");
            } else {
                log.error("发起认证请求失败:" + response.getBody().getMessage());
                throw new CyServiceException("发起认证请求失败");
            }
            return map;
        } else
            throw new CyServiceException("参数错误，请重新发起验证");
    }


    private  InitFaceVerifyResponse initFaceVerifyAutoRoute(InitFaceVerifyRequest request) {
        // 第一个为主区域Endpoint，第二个为备区域Endpoint。
        List<String> endpoints = List.of(faceVerifyAliComp.getEndpoints());
        InitFaceVerifyResponse lastResponse = null;
        for (int i=0; i<endpoints.size(); i++) {
            try {
                InitFaceVerifyResponse response = initFaceVerify(endpoints.get(i), request);
                lastResponse = response;

                // 服务端错误，切换到下个区域调用。
                if(response != null){
                    if(500 == response.getStatusCode()){
                        continue;
                    }
                    if(response.getBody() != null){
                        if("500".equals(response.getBody().getCode())){
                            continue;
                        }
                    }
                }

                // 正常返回
                return lastResponse;
            }catch (Exception e) {
                e.printStackTrace();
                if(i == endpoints.size()-1){
                    throw new RuntimeException(e);
                }
            }
        }

        return lastResponse;
    }

    private InitFaceVerifyResponse initFaceVerify(String endpoint, InitFaceVerifyRequest request)
            throws Exception {
        // 阿里云账号AccessKey拥有所有API的访问权限，建议您使用RAM用户进行API访问或日常运维。
        // 强烈建议不要把AccessKey ID和AccessKey Secret保存到工程代码里，否则可能导致AccessKey泄露，威胁您账号下所有资源的安全。
        // 本示例通过阿里云Credentials工具从环境变量中读取AccessKey，来实现API访问的身份验证。如何配置环境变量，请参见https://help.aliyun.com/document_detail/378657.html。
//        com.aliyun.credentials.Client credentialClient = new com.aliyun.credentials.Client();
        Config config = new Config();
//        config.setCredential(credentialClient);
        config.setEndpoint(endpoint);
        config.setAccessKeyId(faceVerifyAliComp.getAccessKeyId());
        config.setAccessKeySecret(faceVerifyAliComp.getAccessKeySecret());
        // 设置http代理。
        //config.setHttpProxy("http://xx.xx.xx.xx:xxxx");
        // 设置https代理。
        //config.setHttpsProxy("https://xx.xx.xx.xx:xxxx");
        Client client = new Client(config);

        // 创建RuntimeObject实例并设置运行参数。
        RuntimeOptions runtime = new RuntimeOptions();
        runtime.readTimeout = 10000;
        runtime.connectTimeout = 10000;

        return client.initFaceVerifyWithOptions(request, runtime);
    }



    /**
     * 获取认证详细数据
     * @param dto
     * @return
     */
    @Override
    public  Map<String,String> describeFaceVerify(InitFaceVerifyRequest dto) {
        String userId = dto.getUserId();
        if (userId == null)
            userId = CyUserUtil.getAuthenBusinessId();
        Map<String, String> map = new HashMap<>();
        MemInfo memInfo = memInfoRepository.getInfoByUserId(userId);
        if ("0".equals(memInfo.getMemRealAuthen())) {
            // 创建API请求并设置参数。
            DescribeFaceVerifyRequest request = new DescribeFaceVerifyRequest();
            // 场景ID+L。
            request.setSceneId(faceVerifyAliComp.getSceneId());
            // CertifyId在InitFaceVerify接口的返回值中。
            request.setCertifyId(dto.getCertifyId());
            DescribeFaceVerifyResponse response = describeFaceVerifyAutoRoute(request);

            Object detail = cyRedisServiceImpl.get(FaceVerifyRedisBean.FACEVERIFY_USER+userId);
            if (detail!=null) {
                MemInfo info = JSON.parseObject(JSON.toJSONString(detail), MemInfo.class);
                String idCard = info.getMemIdcard();
                String realName = info.getMemRealName();

                log.info(response.getBody().getRequestId());
                log.info(response.getBody().getCode());
                log.info(response.getBody().getMessage());
                map.put("requestId", response.getBody().getRequestId());
                map.put("message", response.getBody().getMessage());
                if ("200".equals(response.getBody().getCode())) {
                    log.info("passed:" + response.getBody().getResultObject().getPassed());
                    log.info("subCode:" + response.getBody().getResultObject().getSubCode());
                    map.put("passed", response.getBody().getResultObject().getPassed());
                    map.put("subCode", response.getBody().getResultObject().getSubCode());
                    if ("T".equals(response.getBody().getResultObject().getPassed())) {
                        memInfo.setMemRealAuthen("1");
                        memInfo.setMemRealAuthenDate(new Date());
                        memInfo.setMemIdcard(CyAESUtils.encrypt(idCard));
                        memInfo.setMemRealName(CyAESUtils.encrypt(realName));
                        memInfoRepository.updateById(memInfo);
                        //修改sys_user
                        SysUserRbac userRbac = sysUserRbacRepository.selectById(memInfo.getUserId());
                        userRbac.setIdNumber(memInfo.getMemIdcard());
                        sysUserRbacRepository.updateById(userRbac);
                    } else {
                        HttpServletRequest request1 = ((ServletRequestAttributes) (RequestContextHolder.getRequestAttributes())).getRequest();
                        String ip = CyAddressUtils.IpUtils.getIpAddr(request1);
                        Boolean ipMin = cyRedisServiceImpl.hasKey(FaceVerifyRedisBean.FACEVERIFY_IP_MINUTES + ip);
                        // 1 设置 ip 每分钟的失败次数
                        if (ipMin) {
                            cyRedisServiceImpl.incr(FaceVerifyRedisBean.FACEVERIFY_IP_MINUTES + ip, 1l);
                        } else
                            cyRedisServiceImpl.set(FaceVerifyRedisBean.FACEVERIFY_IP_MINUTES + ip, 1, 60L);
                        Boolean ipDay = cyRedisServiceImpl.hasKey(FaceVerifyRedisBean.FACEVERIFY_IP_DAY + ip);
                        // 1 设置 ip 每天的失败次数
                        if (ipDay) {
                            cyRedisServiceImpl.incr(FaceVerifyRedisBean.FACEVERIFY_IP_DAY + ip, 1l);
                        } else {
                            long time = TimeUtil.getTimeSecond(new Date(), TimeUtil.getTodayEndTime());
                            cyRedisServiceImpl.set(FaceVerifyRedisBean.FACEVERIFY_IP_DAY + ip, 1, time);
                        }
                        Boolean idCardMin = cyRedisServiceImpl.hasKey(FaceVerifyRedisBean.FACEVERIFY_IDCARD_MINUTES + idCard);
                        // 1 设置 ip 每分钟的失败次数
                        if (idCardMin) {
                            cyRedisServiceImpl.incr(FaceVerifyRedisBean.FACEVERIFY_IDCARD_MINUTES + idCard, 1l);
                        } else
                            cyRedisServiceImpl.set(FaceVerifyRedisBean.FACEVERIFY_IDCARD_MINUTES + idCard, 1, 60L);
                        Boolean idCardDay = cyRedisServiceImpl.hasKey(FaceVerifyRedisBean.FACEVERIFY_IDCARD_DAY + idCard);
                        // 1 设置 ip 每天的失败次数
                        if (idCardDay) {
                            cyRedisServiceImpl.incr(FaceVerifyRedisBean.FACEVERIFY_IDCARD_DAY + idCard, 1l);
                        } else {
                            long time = TimeUtil.getTimeSecond(new Date(), TimeUtil.getTodayEndTime());
                            cyRedisServiceImpl.set(FaceVerifyRedisBean.FACEVERIFY_IDCARD_DAY + idCard, 1, time);
                        }
                    }
                } else {
                    log.error("获取认证详细数据失败:" + response.getBody().getMessage());
                    throw new CyServiceException(response.getBody().getMessage());
                }
            }
        }
        map.put("realAuthen", memInfo.getMemRealAuthen());
        return map;
    }

    private DescribeFaceVerifyResponse describeFaceVerifyAutoRoute(DescribeFaceVerifyRequest request) {
        // 第一个为主区域Endpoint，第二个为备区域Endpoint。
        List<String> endpoints = List.of(faceVerifyAliComp.getEndpoints());
        DescribeFaceVerifyResponse lastResponse = null;
        for (int i = 0; i < endpoints.size(); i++) {
            try {
                DescribeFaceVerifyResponse response = describeFaceVerify(endpoints.get(i), request);
                lastResponse = response;

                // 服务端错误，切换到下个区域调用。
                if (response != null) {
                    if (500 == response.getStatusCode()) {
                        continue;
                    }
                    if (response.getBody() != null) {
                        if ("500".equals(response.getBody().getCode())) {
                            continue;
                        }
                    }
                }

                return lastResponse;
            } catch (Exception e) {
                if (i == endpoints.size() - 1) {
                    throw new RuntimeException(e);
                }
            }
        }

        return lastResponse;
    }

    private DescribeFaceVerifyResponse describeFaceVerify(String endpoint, DescribeFaceVerifyRequest request)
            throws Exception {
        // 阿里云账号AccessKey拥有所有API的访问权限，建议您使用RAM用户进行API访问或日常运维。
        // 强烈建议不要把AccessKey ID和AccessKey Secret保存到工程代码里，否则可能导致AccessKey泄露，威胁您账号下所有资源的安全。
        //本示例通过阿里云Credentials工具从环境变量中读取AccessKey，来实现API访问的身份验证。如何配置环境变量，请参见https://help.aliyun.com/document_detail/378657.html。
//        com.aliyun.credentials.Client credentialClient = new com.aliyun.credentials.Client();
        Config config = new Config();
//        config.setCredential(credentialClient);
        config.setEndpoint(endpoint);
        config.setAccessKeyId(faceVerifyAliComp.getAccessKeyId());
        config.setAccessKeySecret(faceVerifyAliComp.getAccessKeySecret());
        Client client = new Client(config);

        RuntimeOptions runtime = new RuntimeOptions();
        runtime.readTimeout = 10000;
        runtime.connectTimeout = 10000;

        return client.describeFaceVerifyWithOptions(request, runtime);
    }



    @Override
    public void callbackFaceVerify(String callbackToken, String certifyId, String passed) {
//        String userId = CyUserUtil.getUserInfoStrProp(callbackToken,CyUserUtil.USER_ID);
        String userId = CyAESUtils.decrypt(callbackToken);
        if (userId!=null){
            log.info("回调certifyId:"+certifyId);
            log.info("回调passed:"+passed);
            MemInfo memInfo = memInfoRepository.getInfoByUserId(userId);
            //认证成功
            if ("0".equals(memInfo.getMemRealAuthen())) {
                Object detail = cyRedisServiceImpl.get(FaceVerifyRedisBean.FACEVERIFY_USER + userId);
                if (detail!=null) {
                    MemInfo info = JSON.parseObject(JSON.toJSONString(detail), MemInfo.class);
                    String idCard = info.getMemIdcard();
                    String realName = info.getMemRealName();
                    if ("200".equals(passed)) {
                        memInfo.setMemRealAuthen("1");
                        memInfo.setMemRealAuthenDate(new Date());
                        memInfo.setMemIdcard(CyAESUtils.encrypt(idCard));
                        memInfo.setMemRealName(CyAESUtils.encrypt(realName));
                        memInfoRepository.updateById(memInfo);
                        //修改sys_user
                        SysUserRbac userRbac = sysUserRbacRepository.selectById(memInfo.getUserId());
                        userRbac.setIdNumber(memInfo.getMemIdcard());
                        sysUserRbacRepository.updateById(userRbac);
                    } else {
                        HttpServletRequest request1 = ((ServletRequestAttributes) (RequestContextHolder.getRequestAttributes())).getRequest();
                        String ip = CyAddressUtils.IpUtils.getIpAddr(request1);
                        Boolean ipMin = cyRedisServiceImpl.hasKey(FaceVerifyRedisBean.FACEVERIFY_IP_MINUTES + ip);
                        // 1 设置 ip 每分钟的失败次数
                        if (ipMin) {
                            cyRedisServiceImpl.incr(FaceVerifyRedisBean.FACEVERIFY_IP_MINUTES + ip, 1l);
                        } else
                            cyRedisServiceImpl.set(FaceVerifyRedisBean.FACEVERIFY_IP_MINUTES + ip, 1, 60L);
                        Boolean ipDay = cyRedisServiceImpl.hasKey(FaceVerifyRedisBean.FACEVERIFY_IP_DAY + ip);
                        // 1 设置 ip 每天的失败次数
                        if (ipDay) {
                            cyRedisServiceImpl.incr(FaceVerifyRedisBean.FACEVERIFY_IP_DAY + ip, 1l);
                        } else {
                            long time = TimeUtil.getTimeSecond(new Date(), TimeUtil.getTodayEndTime());
                            cyRedisServiceImpl.set(FaceVerifyRedisBean.FACEVERIFY_IP_DAY + ip, 1, time);
                        }
                        Boolean idCardMin = cyRedisServiceImpl.hasKey(FaceVerifyRedisBean.FACEVERIFY_IDCARD_MINUTES + idCard);
                        // 1 设置 ip 每分钟的失败次数
                        if (idCardMin) {
                            cyRedisServiceImpl.incr(FaceVerifyRedisBean.FACEVERIFY_IDCARD_MINUTES + idCard, 1l);
                        } else
                            cyRedisServiceImpl.set(FaceVerifyRedisBean.FACEVERIFY_IDCARD_MINUTES + idCard, 1, 60L);
                        Boolean idCardDay = cyRedisServiceImpl.hasKey(FaceVerifyRedisBean.FACEVERIFY_IDCARD_DAY + idCard);
                        // 1 设置 ip 每天的失败次数
                        if (idCardDay) {
                            cyRedisServiceImpl.incr(FaceVerifyRedisBean.FACEVERIFY_IDCARD_DAY + idCard, 1l);
                        } else {
                            long time = TimeUtil.getTimeSecond(new Date(), TimeUtil.getTodayEndTime());
                            cyRedisServiceImpl.set(FaceVerifyRedisBean.FACEVERIFY_IDCARD_DAY + idCard, 1, time);
                        }
                    }
                }
            }
        }
    }
}
